/*
 *	main.c
 *
 *	USB-LAN (pingv)
 *
 *	* Sat Aug 27 17:04:00 JST 2005 Naoyuki Sawa
 *	- 1st [XB
 */
#include "app.h"

/****************************************************************************
 *	
 ****************************************************************************/

#define extern
#include "work.h"
#undef extern

/****************************************************************************
 *	main
 ****************************************************************************/

int
main()
{
	/* B */
	app_init();

	/* C[vB */
	app_main();

	return 0;
}

/****************************************************************************
 *	
 ****************************************************************************/

#define SAVE_FNAME		"usblan.sav"
#define INIT_IP_ADDRESS		IP_ADDRESS(192,168,255,  1)
#define INIT_SUBNET_MASK	IP_ADDRESS(255,255,255,  0)
#define INIT_DEFAULT_GATEWAY	IP_ADDRESS(192,168,255,254)
#define INIT_PING_ADDRESS	IP_ADDRESS(192,168,255,254)

typedef struct _CONFIG {
	IPCONFIG ip_config;	/* lbg[Nݒ */
	int ping_address;	/* pingĐAhX */
} CONFIG;

static void
load_config(CONFIG* c)
{
	int load_ok = 0;
	//
	int retval;
	FILE* fp;

	/* Z[ut@CAݒǂݍ݂܂B */
	fp = fopen(SAVE_FNAME, "rb");
	if(fp) {
		retval = fread(c, sizeof(CONFIG), 1, fp);
		if(retval == 1) {
			load_ok = 1;
		}
		fclose(fp);
	}

	/* ǂݍݎsȂ΁Ali[܂B */
	if(!load_ok) {
		c->ip_config.ip_address      = INIT_IP_ADDRESS;
		c->ip_config.subnet_mask     = INIT_SUBNET_MASK;
		c->ip_config.default_gateway = INIT_DEFAULT_GATEWAY;
		c->ping_address              = INIT_PING_ADDRESS;
	}
}

static void
save_config(const CONFIG* c)
{
	int retval;
	FILEACC fa;

	/* ݂̃Z[ut@CeA܂́Al擾܂B */
	CONFIG cc;
	load_config(&cc);

	/* ݒύXA܂́Al̂܂܂Ȃ΁AZ[u܂B */
	if(memcmp(c, &cc, sizeof(CONFIG)) == 0) {
		return;
	}

	/* lbg[NݒAZ[ut@Cɏ݂܂B */
	retval = pceFileCreate(SAVE_FNAME, sizeof(CONFIG));
	if(retval == 0) {
		retval = pceFileOpen(&fa, SAVE_FNAME, FOMD_WR);
		if(retval == 0) {
			pceFileWriteSct(&fa, c, 0, sizeof(CONFIG));
			pceFileClose(&fa);
		}
	}
}

/****************************************************************************
 *	app_init
 ****************************************************************************/

void
app_init()
{
	CONFIG c;

	/* ʓIȏB */
	pceAppSetProcPeriod(1000 / FPS);
	pceLCDSetBuffer(vbuff);
	pceLCDDispStart();
	pcePadSetTrigMode(PP_MODE_SINGLE);

	/* `B */
	surface.w = DISP_X;
	surface.h = DISP_Y;
	surface.vbuff = vbuff;
	render_init_2d(&render, &surface, 0, 0, 0, 0);

	/* N[Abv֐o^B */
	atexit(app_exit);

	/* PCrunŎsꍇUSB̂߂ɁAxď܂B */
	delay(SEC(0.1));
	ethernet_start(0);
	ip_start();
	pceTimerSetCallback(0, PCE_TT_PERIODIC, 5 * 1000, (void*)arp_gratuitous); /* ΏǗÖ@ */

	/* ݒ[h܂B */
	load_config(&c);
	ip_set_config(&c.ip_config);
	ping_address = c.ping_address;
	initialized = 1;
}

/****************************************************************************
 *	app_exit
 ****************************************************************************/

void
app_exit()
{
	CONFIG c;

	/* ݒZ[u܂B */
	if(initialized) {
		ip_get_config(&c.ip_config);
		c.ping_address = ping_address;
		save_config(&c);
	}

	/* mɐؒf܂B */
	pceTimerSetCallback(0, PCE_TT_NONE, 0, NULL); /* ΏǗÖ@ */
	ip_stop();
	ethernet_stop();
}

/****************************************************************************
 *	schedule
 ****************************************************************************/

void
schedule()
{
	static unsigned char rep[8]; /* I[gs[gJE^ */
	//
	int x;
	int y;

	/* ʍXVB */
	pceLCDTrans();

	/* VXeB */
	yield();

	/* pbhԍXVB */
	joy = pcePadGet();

	/* I[gs[gB */
	for(x = 0; x < 8; x++) {
		y = 1 << x; /* PAD_* */
		if(QUICK_REPEAT_PADS & y) {
			if(joy & y) {
				if(rep[x] != 0xff) rep[x]++;
				if(rep[x] >= QUICK_REPEAT_DELAY) joy |= y << 8; /* TRG_* */
			} else {
				rep[x] = 0;
			}
		}
	}

	/* O[o^C}XVB */
	now++;

	/* SELECT{^ꂽAI܂B */
	if(joy & TRG_SELECT) {
		exit(0);
	}

	/* J[\̐Fݒ肵܂B */
	if(joy) {
		cursor_count = 0;
	} else {
		cursor_count++;
	}
	if(cursor_count >= SEC(0.5)) {
		cursor_count = 0;
	}
	if(cursor_count < SEC(0.3)) {
		cursor_blink = 1;
	} else {
		cursor_blink = 0;
	}
}

/****************************************************************************
 *	delay
 ****************************************************************************/

void
delay(int frame)
{
	while(frame > 0) {
		schedule();
		frame--;
	}
}

/****************************************************************************
 *	app_main
 ****************************************************************************/

static void do_ip_config();
static void do_icmp_echo();
static void do_ethernet_stat();
static void do_arp_table();

void
app_main()
{
	for(;;) {
		do_ip_config();
		do_icmp_echo();
		do_ethernet_stat();
		do_arp_table();
	}
}

static void
init_window(const char* title)
{
	surface_clear(&surface, 0);

	render_rectangle_fill(&render, 0, 0, 128, 10, 3);
	render_printf(&render, 0, 0, 0x80, 0, "%s", title);
}

/*	lbg[Nݒ
 *	IPڽ:
 *	  XXX.XXX.XXX.XXX Kp
 *	ȯϽ:
 *	  XXX.XXX.XXX.XXX
 *	̫Ĺްĳ:
 *	  XXX.XXX.XXX.XXX
 */
static void
do_ip_config()
{
	int i;
	int j;
	int x;
	int y;
	int col;
	int row;
	int modify;
	int apply;
	int fore;
	int back;
	IPCONFIG ic;
	unsigned char ia[3][4];

	for(;;) {
		col    = 3;
		row    = 0;
		modify = 0;
		apply  = 0;

		/* lbg[Nݒ擾܂B */
		ip_get_config(&ic);

		/* lbg[NݒWJ܂B */
		PUT_BEWORD(ia[0], ic.ip_address);
		PUT_BEWORD(ia[1], ic.subnet_mask);
		PUT_BEWORD(ia[2], ic.default_gateway);

		do {
			schedule();
			if(joy & TRG_START) {
				return;
			}
			init_window("lbg[Nݒ");

			/* ݒ荀ڂ̑IAݒl̑AKps܂B */
			if(col < 4) {
				if((joy & TRG_UP) && (row > 0)) {
					row--;
				}
				if((joy & TRG_DN) && (row < 2)) {
					row++;
				}
			}
			if((joy & TRG_LF) && (col > 0)) {
				col--;
			}
			if((joy & TRG_RI) && (col < (modify ? 4 : 3))) {
				col++;
			}
			if(col < 4) {
				if(joy & TRG_A) {
					ia[row][col]++;
					if(row == 0) {
						/* IPAhXύXꂽATulbg}XNݒ肵܂B */
						PUT_BEWORD(ia[1], address_class_mask(BEWORD(ia[0])));
					}
					modify = 1;
				}
				if(joy & TRG_B) {
					ia[row][col]--;
					if(row == 0) {
						/* IPAhXύXꂽATulbg}XNݒ肵܂B */
						PUT_BEWORD(ia[1], address_class_mask(BEWORD(ia[0])));
					}
					modify = 1;
				}
			} else {
				if(joy & (TRG_A | TRG_B)) {
					/* ʕ\ɁAlbg[Nݒ̓KpցB */
					apply = 1;
				}
			}

			/* ʕ\s܂B */
			y = 11;
			for(i = 0; i < 3; i++) {
				x = 2;
				switch(i) {
				case 0:
					render_printf(&render, x, y, 0x80, 3, "IPڽ:");
					break;
				case 1:
					render_printf(&render, x, y, 0x80, 3, "ȯϽ:");
					break;
				case 2:
					render_printf(&render, x, y, 0x80, 3, "̫Ĺްĳ:");
					break;
				}
				y += 10;

				x = 2 + 10;
				for(j = 0; j < 4; j++) {
					if((i == row) && (j == col) && cursor_blink) {
						fore = 0;
						back = 3;
					} else {
						fore = 3;
						back = 1;
					}
					render_rectangle_fill(&render, x, y, 5 * 3, 10, back);
					render_printf(&render, x, y, 0x80, fore, "%3d", ia[i][j]);
					x += 5 * 3;
					if(j < 3) {
						render_printf(&render, x, y, 0x80, 3, ".");
					}
					x += 5;
				}
				if(i == 0) {
					if((j == col) && cursor_blink) {
						fore = 0;
						back = 3;
					} else {
						fore = modify ? 3 : 2;
						back = 1;
					}
					render_rectangle_fill(&render, x, y, 5 * 4, 10, back);
					render_printf(&render, x, y, 0x80, fore, "Kp");
				}
				y += 10;
			}
		} while(!apply);

		/* lbg[Nݒ̓WJ߂܂B */
		ic.ip_address      = BEWORD(ia[0]);
		ic.subnet_mask     = BEWORD(ia[1]);
		ic.default_gateway = BEWORD(ia[2]);

		/* lbg[NݒKp܂B */
		ip_set_config(&ic);
	}
}

/*	
 *	Đڽ:
 *	  XXX.XXX.XXX.XXX s
 *	:
 *	   time=Xms
 *	M:     X
 *	vM:     X
 */
static void
do_icmp_echo()
{
	static int elapse = INT_MAX;
	//
	int col = 4;
	//
	int i;
	int x;
	int y;
	int run;
	int fore;
	int back;
	unsigned char ia[4];
	ICMPSTAT is;

	for(;;) {
		schedule();
		if(joy & TRG_START) {
			return;
		}
		init_window("");

		run = 0;

		/* ĐAhXWJ܂B */
		PUT_BEWORD(ia, ping_address);

		/* ݒ荀ڂ̑IAݒl̑s܂B */
		if((joy & TRG_LF) && (col > 0)) {
			col--;
		}
		if((joy & TRG_RI) && (col < 4)) {
			col++;
		}
		if(col < 4) {
			if(joy & TRG_A) {
				ia[col]++;
				elapse = INT_MAX; /* ĐAhXύXAʂNA */
			}
			if(joy & TRG_B) {
				ia[col]--;
				elapse = INT_MAX; /* ĐAhXύXAʂNA */
			}
		} else {
			if(joy & (TRG_A | TRG_B)) {
				run = 1; /* ʕ\ɁApings */
			}
		}

		/* ĐAhX̓WJ߂܂B */
		ping_address = BEWORD(ia);

		/* ICMPʐMԂ擾܂B */
		icmp_stat(&is);

		/* ʕ\s܂B */
		y = 11;
		x = 2;
		render_printf(&render, x, y, 0x80, 3, "Đڽ:");
		y += 10;
		x = 2 + 10;
		for(i = 0; i < 5; i++) {
			if((i == col) && cursor_blink) {
				fore = 0;
				back = 3;
			} else {
				fore = 3;
				back = 1;
			}
			if(i < 4) {
				render_rectangle_fill(&render, x, y, 5 * 3, 10, back);
				render_printf(&render, x, y, 0x80, fore, "%3d", ia[i]);
				x += 5 * 3;
				if(i < 3) {
					render_printf(&render, x, y, 0x80, 3, ".");
				}
			} else {
				render_rectangle_fill(&render, x, y, 5 * 4, 10, back);
				render_printf(&render, x, y, 0x80, fore, "s");
			}
			x += 5;
		}
		y += 10;
		x = 2;
		render_printf(&render, x, y, 0x80, 3, ":");
		y += 10;
		x = 2 + 10;
		if(!run) {
			if(elapse != INT_MAX) {
				if(elapse >= 0) {
					if(!elapse) {
						render_printf(&render, x, y, 0x80, 3, " time<1ms");
					} else {
						render_printf(&render, x, y, 0x80, 3, " time=%dms", elapse);
					}
				} else {
					render_printf(&render, x, y, 0x80, 3, "s");
				}
			}
		} else {
			render_printf(&render, x, y, 0x80, 3, "s...");
		}
		y += 10;
		x = 2;
		render_printf(&render, x, y, 0x80, 3, "M:%6d", is.echo_reply);
		y += 10;
		x = 2;
		render_printf(&render, x, y, 0x80, 3, "vM:%6d", is.echo_request);

		/* pings܂B */
		if(run) {
			schedule(); /* ʍXV */
			elapse = icmp_echo_request(ping_address, -1);
			run = 0;
		}
	}
}

/*	k`mʐM
 *	MACڽ:
 *	  XX-XX-XX-XX-XX-XX
 *	M:         X
 *	Ms:         X
 *	M:         X
 *	Ms:         X
 */
static void
do_ethernet_stat()
{
	int x;
	int y;
	unsigned char ea[6];
	ETHERNETSTAT es;

	for(;;) {
		schedule();
		if(joy & TRG_START) {
			return;
		}
		init_window("k`mʐM");

		/* MACAhX擾܂B */
		ethernet_get_address(ea);

		/* ʐMԂ擾܂B */
		ethernet_stat(&es);

		/* ʕ\s܂B */
		y = 11;
		x = 2;
		render_printf(&render, x, y, 0x80, 3, "MACڽ:");
		y += 10;
		x = 2 + 10;
		render_printf(&render, x, y, 0x80, 3, "%02x-%02x-%02x-%02x-%02x-%02x", ea[0],ea[1],ea[2],ea[3],ea[4],ea[5]);
		y += 10;
		x = 2;
		render_printf(&render, x, y, 0x80, 3, "Mڰ:%6d", es.rxok);
		y += 10;
		render_printf(&render, x, y, 0x80, 3, "Msڰ:%6d", es.rxerr);
		y += 10;
		render_printf(&render, x, y, 0x80, 3, "Mڰ:%6d", es.txok);
		y += 10;
		render_printf(&render, x, y, 0x80, 3, "Msڰ:%6d", es.txerr);
		y += 10;
		render_printf(&render, x, y, 0x80, 3, "BusReset:%6d", es.reset);
	}
}

/*	`qoe[u
 *	XXX.XXX.XXX.XXX : XX-XX-XX-XX-XX-XX
 *	XXX.XXX.XXX.XXX : XX-XX-XX-XX-XX-XX
 *	XXX.XXX.XXX.XXX : XX-XX-XX-XX-XX-XX
 *	XXX.XXX.XXX.XXX : XX-XX-XX-XX-XX-XX
 */
static void
do_arp_table()
{
	int i;
	int j;
	int n;
	int x;
	int y;
	ARPENTRY at[88 / 6];
	unsigned char ia[4];
	unsigned char ea[6];

	for(;;) {
		schedule();
		if(joy & TRG_START) {
			return;
		}
		init_window("`qoe[u");

		/* ARPe[u擾܂B */
		n = arp_table(at, ARRAY_SIZE(at));

		/* ʕ\s܂B */
		y = 11;
		if(!n) {
			x = 2;
			render_printf(&render, x, y, 0x80, 3, "o^Ă܂");
			continue;
		}
		for(i = 0; i < n; i++) {
			PUT_BEWORD(ia, at[i].ip_address);
			x = 2;
			for(j = 0; j < 4; j++) {
				render_printf(&render, x, y, 0x82, 3, "%3d", ia[j]);
				x += 4 * 3;
				if(j < 3) {
					render_point(&render, x, y + 4, 3);
					x += 2;
				}
			}

			x += 2;
			render_printf(&render, x, y, 0x82, 3, ":");
			x += 4;

			memcpy(ea, at[i].ethernet_address, 6);
			x += 2;
			for(j = 0; j < 6; j++) {
				render_printf(&render, x, y, 0x82, 3, "%02X", ea[j]);
				x += 4 * 2;
				if(j < 5) {
					render_point(&render, x + 0, y + 2, 3);
					render_point(&render, x + 1, y + 2, 3);
					x += 3;
				}
			}
			y += 6;
		}
	}
}

